#include "sxConsole.h"
#include "sxTypes.h"

#define VGA_CRT_ADDR_REG 0x3D4
#define VGA_CRT_DATA_REG 0x3D5
#define VGA_CRT_CURSOR_LOC_HIGH_REG 0x0E
#define VGA_CRT_CURSOR_LOC_LOW_REG 0x0F

static sx_uint32 SxConWidth			= 80;
static sx_uint32 SxConHeight		= 24; /* Core Status line taken into account already */
static sx_uint32 SxConPosX			= 0;
static sx_uint32 SxConPosY			= 0;
static sx_uint16 SxConAttribute		= 7;
static sx_uint8  SxConCursorVisible	= 1;
static sx_uint16* SxConMemory		= (sx_uint16*)0xB8000;

static void SxConScroll() {
	sx_uint32 t;

	for(t = 0; t < SxConWidth * (SxConHeight - 1); t++)
		*(SxConMemory + t) = *(SxConMemory + t + SxConWidth);
	for(; t < SxConWidth * SxConHeight; t++)
		*(SxConMemory + t) = ' ' | (SxConAttribute << 8);
}

static void SxConSetCursor() {
	if(SxConCursorVisible) {
		sx_uint32 charoff;
		sx_uint8 origaddr;

		/* Compute new cursor location (character offset) */
		charoff = (SxConPosY * SxConWidth) + SxConPosX;

		/* get original contents of VGA address register */
		origaddr = _inp(VGA_CRT_ADDR_REG);

		/* move the cursor */
		_outp(VGA_CRT_ADDR_REG, VGA_CRT_CURSOR_LOC_HIGH_REG);
		_outp(VGA_CRT_DATA_REG, (charoff >> 8) & 0xFF);
		_outp(VGA_CRT_ADDR_REG, VGA_CRT_CURSOR_LOC_LOW_REG);
		_outp(VGA_CRT_DATA_REG, charoff & 0xFF);

		/* restore contents of VGA address register */
		_outp(VGA_CRT_ADDR_REG, origaddr);
	}
}

static void SxConNewLine() {
	SxConPosX = 0;

	if(SxConPosY == SxConHeight - 1)
		SxConScroll();
	else
		SxConPosY++;

	SxConSetCursor();
}

static void SxConNumToStr(char *buf, sx_int32 base, sx_int32 d) {
	char *p = buf;
	char *p1, *p2;
	char off = 'a';
	sx_uint32 ud = d;
	sx_int32 divisor = 10;

	if(base == 'X' || base == 'P')
	{
		base = (base == 'X' ? 'x' : 'p');
		off = 'A';
	}
	
	/* If %d is specified and D is minus, put - in the head. */
	if (base == 'd' && d < 0)
	{
		*p++ = '-';
		buf++;
		ud = -d;
	}
	else if (base == 'x' || base == 'p')
		divisor = 16;

	/* Divide UD by DIVISOR until UD == 0. */
	do {
		int remainder = ud % divisor;
		*p++ = (remainder < 10) ? remainder + '0' : remainder + off - 10;
	} while (ud /= divisor);

	if(base == 'p' || base == 'P') {
		/* add zeros until string is 8 chars */
		while(((sx_uint32)p - (sx_uint32)buf) < 8) *p++='0';
	}

	/* Terminate BUF. */
	*p = 0;

	/* Reverse BUF. */
	p1 = buf;
	p2 = p - 1;

	while (p1 < p2)
	{
		char tmp = *p1;
		*p1 = *p2;
		*p2 = tmp;
		p1++;
		p2--;
	}
}

void SxConClear() {
	sx_uint32 i;

	for(i = 0; i < SxConWidth * SxConHeight; ++i)
		*(SxConMemory + i) = 0 | SxConAttribute << 8;

	SxConPosX = 0;
	SxConPosY = 0;

	SxConSetCursor();
}

void SxConClearLine() {
	sx_uint32 x;
	
	for(x = 0; x < SxConWidth; ++x)
		*(SxConMemory + (x + SxConPosY * SxConWidth)) = 0 | SxConAttribute << 8;
}

void SxConChar(char c) {
	if(c == '\n' || c == '\r') {
		SxConNewLine();
		return;
	}

	*(SxConMemory + (SxConPosX + SxConPosY * SxConWidth)) = c | SxConAttribute << 8;

	SxConPosX++;

	if(SxConPosX >= SxConWidth)
		SxConNewLine();

	SxConSetCursor();
}

void SxConPrintVa(char const * fmt, sx_va_list va) {
	char c;

	SxConCursorVisible = 0;

	while((c = *fmt++) != 0) {
		if(c != '%') {
			SxConChar(c);
		} else {
			c = *fmt++;

			switch(c) {
				case 'c':
					SxConChar(sx_va_arg(va, char));
					break;
				case 's':
					{
						char const* str = sx_va_arg(va, char const*);

						if(!str) str = "(null)";
						while(*str) SxConChar(*str++);
					}
					break;
				case 'p':
				case 'P':
					SxConPrint("0x");
				case 'x':
				case 'X':
				case 'd':
				case 'u':
					{
						char const* str;
						char buf[32];

						SxConNumToStr(buf, c, sx_va_arg(va, sx_uint32));

						str = buf;
						while(*str) SxConChar(*str++);
					}
					break;
				case 'b':
					{
						sx_uint32 i;
						sx_uint32 l = 32;
						sx_uint32 v = sx_va_arg(va, sx_uint32);

						c = *fmt++;

						switch(c) {
							case 'b': l = 8;  break;
							case 'w': l = 16; break;
							case 'd': l = 32; break;
							case 'q': l = 64; break;
							default:
								c = *--fmt;
						}

						for(i = l; i > 0; --i) {
							SxConChar('0' + ((v >> (i - 1)) & 0x1));
						}
					}
					break;
				default:
					SxConPrint("<?>");
					break;
			}
		}
	}

	sx_va_end(va);

	SxConCursorVisible = 1;
	SxConSetCursor();
}

void SxConPrint(char const * fmt, ...) {
	sx_va_list va;
	sx_va_start(va, fmt);

	SxConPrintVa(fmt, va);

	sx_va_end(va);
}

void SxConSetBackColor(sx_eColors col) {
	SxConAttribute = (SxConAttribute & 0x0F) | (char)col << 4;
}

void SxConSetTextColor(sx_eColors col) {
	SxConAttribute = (SxConAttribute & 0xF0) | (char)col;
}

void SxConSetColor(sx_eColors fore, sx_eColors back) {
	SxConAttribute = (SxConAttribute & 0x0F) | (char)back << 4;
	SxConAttribute = (SxConAttribute & 0xF0) | (char)fore;
}

void SxConGetColor(sx_eColors* fore, sx_eColors* back) {
	if(fore) *fore = (sx_eColors)(SxConAttribute & 0x0F);
	if(back) *back = (sx_eColors)((SxConAttribute >> 4) & 0x0F);
}

void SxConSetPos(sx_uint32 x, sx_uint32 y) {
	SxConPosX = x;
	SxConPosY = y;
	SxConSetCursor();
}

void SxConGetPos(sx_uint32* x, sx_uint32 *y) {
	if(x) *x = SxConPosX;
	if(y) *y = SxConPosY;
}

void SxConGetSize(sx_uint32* w, sx_uint32* h) {
	if(w) *w = SxConWidth;
	if(h) *h = SxConHeight;
}
