#include "sxInterupts.h"
#include "sxSegmentation.h"
#include "sxConsole.h"
#include "sxMain.h"

#define SX_INT_TYPE_TASK	0x5
#define SX_INT_TYPE_INT		0x6
#define SX_INT_TYPE_TRAP	0x7

#pragma pack(push, 1)

typedef struct {
	sx_uint16 limit;
	sx_uint32 base;
} sx_idtptr_t;

typedef struct {
	sx_uint16	base_low;
	sx_uint16	selector;
	sx_uint8	reserved0;
	sx_uint8	type		: 3;
	sx_uint8	size		: 1;
	sx_uint8	reserved1	: 1;
	sx_uint8	dpl			: 2;
	sx_uint8	present		: 1;
	sx_uint16	base_high;
} sx_interupt_gate_t;

typedef union {
	sx_interupt_gate_t	interupt;
} sx_idt_gate_t;

#pragma pack(pop)

static SX_ALIGN(8) sx_idt_gate_t	sxIdtGates[SX_INTERUPT_COUNT];
static SX_ALIGN(8) sx_idtptr_t		sxIdt = { (sizeof(sx_idt_gate_t) * SX_INTERUPT_COUNT) - 1, (sx_uint32)&sxIdtGates };

void SxInteruptsInit() {
	sx_uint8 master_mask, slave_mask;

	__asm {
		lidt	sxIdt
	}

	SxConPrint(" * Interupt Desciptor Table loaded from %p\n", &sxIdt);

	master_mask = _inp(SX_PIC_MASTER_DAT);
	slave_mask  = _inp(SX_PIC_SLAVE_DAT);

	/* begin init, include ICW4 */
	_outp(SX_PIC_MASTER_CMD,	SX_PIC_ICW1_INIT | SX_PIC_ICW1_IC4);
	_outp(SX_PIC_SLAVE_CMD,		SX_PIC_ICW1_INIT | SX_PIC_ICW1_IC4);

	/* set the interupt vector offsets */
	_outp(SX_PIC_MASTER_DAT,	SX_IRQ_BEGIN);
	_outp(SX_PIC_SLAVE_DAT,		SX_IRQ_BEGIN + 8);

	/* now send the IRQ lines to connect the two PICs */
	_outp(SX_PIC_MASTER_DAT,	4);
	_outp(SX_PIC_SLAVE_DAT,		2);

	/* now ICW4 */
	_outp(SX_PIC_MASTER_DAT,	SX_PIC_ICW4_8086);
	_outp(SX_PIC_SLAVE_DAT,		SX_PIC_ICW4_8086);

	SxConPrint(" * IRQ's remapped to higher Interupts ...\n");
	SxConPrint(" * Interupt Mask on PICs: Master: %bb Slave: %bb\n", master_mask, slave_mask);

	_outp(SX_PIC_MASTER_DAT,	master_mask);
	_outp(SX_PIC_SLAVE_DAT,		slave_mask);
}

void SxMaskPICInterupt(sx_uint8 idx) {
	sx_uint8 mask;

	if(idx > 15)
		SxCorePanic("IRQ Index out of bounds: %d", idx);

	if(idx < 8)
		mask = _inp(SX_PIC_MASTER_DAT);
	else
		mask = _inp(SX_PIC_SLAVE_DAT);

	if(!(mask & (1 << idx))) {
		mask |= (1 << idx);
		SxConPrint(" * Masked IRQ%d on %s PIC, new mask: %bb\n", idx, (idx < 8 ? "master" : "slave"), mask);
	} else {
		return;
	}

	if(idx < 8)
		_outp(SX_PIC_MASTER_DAT, mask);
	else
		_outp(SX_PIC_SLAVE_DAT, mask);
}

void SxUnmaskPICInterupt(sx_uint8 idx) {
	sx_uint8 mask;

	if(idx > 15)
		SxCorePanic("IRQ Index out of bounds: %d", idx);

	if(idx < 8)
		mask = _inp(SX_PIC_MASTER_DAT);
	else
		mask = _inp(SX_PIC_SLAVE_DAT);

	if(mask & (1 << idx)) {
		mask ^= (1 << idx);
		SxConPrint(" * Unmasked IRQ%d on %s PIC, new mask: %bb\n", idx, (idx < 8 ? "master" : "slave"), mask);
	} else {
		return;
	}

	if(idx < 8)
		_outp(SX_PIC_MASTER_DAT, mask);
	else
		_outp(SX_PIC_SLAVE_DAT, mask);
}

void SxMaskAllPICInterupts() {
	_outp(SX_PIC_MASTER_DAT, 0xFF);
	_outp(SX_PIC_SLAVE_DAT, 0xFF);

	SxConPrint(" * All IRQ lines disabled on PICs\n");
}

void SxUnmaskAllPICInterupts() {
	_outp(SX_PIC_MASTER_DAT, 0x0);
	_outp(SX_PIC_SLAVE_DAT, 0x0);

	SxConPrint(" * All IRQ lines enabled on PICs\n");
}

void SxInteruptsEnable() {
	_enable();
}

void SxInteruptsDisable() {
	_disable();
}

void SxIntSetInteruptGate(sx_uint8 idx, sx_uint32 base, sx_uint16 sel, sx_uint8 dpl) {
	sxIdtGates[idx].interupt.base_low = base & 0xFFFF;
	sxIdtGates[idx].interupt.base_high = (base >> 16);

	sxIdtGates[idx].interupt.selector = sel;
	sxIdtGates[idx].interupt.reserved0 = 0;
	sxIdtGates[idx].interupt.type = SX_INT_TYPE_INT;
	sxIdtGates[idx].interupt.size = 1; /* 32 bits */
	sxIdtGates[idx].interupt.reserved1 = 0;
	sxIdtGates[idx].interupt.dpl = dpl;
	sxIdtGates[idx].interupt.present = 1;
}

void SxIntClearAnyGate(sx_uint8 idx) {
	/* no memory management yet, so clear manually */
	sx_uint32* ptr = (sx_uint32*)&sxIdtGates[idx];
	*ptr++	= 0;
	*ptr	= 0;
}
