#include "sxMain.h"
#include "sxBoot.h"
#include "sxTypes.h"
#include "sxConsole.h"
#include "sxInterupts.h"
#include "sxIntHandlers.h"
#include "SxSegmentation.h"

#include "sxDebug.h"
#include "sxTimer.h"

void SxCoreMain(unsigned long magic, sx_multiboot_info* mbi) {
	SxConSetColor(ecBrightWhite, ecBlue);
	SxConClear();

	SxCoreStatus("Setting up Global Descriptor Table ...");
	SxSegmentationInit();

	SxCoreStatus("Setting up Interupt Handlers ...");
	SxIntHandlersInit();

	SxCoreStatus("Initializing Interupts ...");
	SxInteruptsInit();

	SxCoreStatus("Enabling Interupts ...");

	SxCoreStatus("Checking Multiboot Information ...");

	if(magic != SX_BOOT_MAGIC) {
		SxConPrint("Warning: Bad boot magic number: 0x%x\n", magic);
	}

	if(!mbi) {
		SxConPrint("ERROR: Missing multiboot Header\n");
		SxCoreHalt();
	}

	SxConPrint(" * Multiboot Flags: %b\n", mbi->flags);

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_MEMINFO)) {
		SxConPrint(" * Memory: Lower: %uKB, Upper: %uKB\n", mbi->mem_lower, mbi->mem_upper);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_BOOTDEV)) {
		sx_uint8* p = (sx_uint8*)&mbi->boot_device;

		SxConPrint(" * Boot Device: 0x%x (", mbi->boot_device);

		/* TODO: those are just some guessed values... */

		switch(p[3]) {
		case 0x00:
			SxConPrint("Floppy");
			break;
		case 0x20:
			SxConPrint("Network");
			break;
		case 0x80:
			SxConPrint("Harddisk");
			break;
		default:
			SxConPrint("Unknown Device: 0x%x", p[3]);
			break;
		}

		switch(p[2]) {
			case 0xFF:
				break;
			default:
				SxConPrint(", Part. Type: 0x%x", p[2]);
				break;
		}

		switch(p[1]) {
			case 0xFF:
				break;
			default:
				SxConPrint(", Subpart. Type: 0x%x", p[1]);
				break;
		}

		switch(p[0]) {
			case 0xFF:
				break;
			default:
				SxConPrint(", Subpart. Type: 0x%x", p[0]);
				break;
		}

		SxConPrint(")\n");
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_CMDLINE)) {
		SxConPrint(" * Command Line: %s\n", mbi->cmdline);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_MODULES)) {
		sx_multiboot_module *mod;
		sx_uint32 i;

		SxConPrint(" * Module Count: %d, at %p\n", mbi->mods_count, mbi->mods_addr);
		
		for(i = 0, mod = (sx_multiboot_module*) mbi->mods_addr; i < mbi->mods_count; i++, mod++)
			SxConPrint("    * start = 0x%x, end = 0x%x, string = %s\n",
				mod->mod_start, mod->mod_end, mod->string);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_AOUT_HEADER) || SX_BITTEST(mbi->flags, SX_BOOTFLAG_ELF_HEADER)) {
		SxConPrint("ERROR: a.out or ELF headers found in kernel!\n");
		SxCoreHalt();
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_MEMMAP)) {
		sx_multiboot_memory_map* mmap;

		SxConPrint(" * MemoryMap: %p, length %u Bytes\n", mbi->mmap_addr, mbi->mmap_length);

		for (mmap = (sx_multiboot_memory_map*)mbi->mmap_addr
			; (sx_uint32)mmap < mbi->mmap_addr + mbi->mmap_length
			; mmap = (sx_multiboot_memory_map*)((sx_uint32) mmap 
				+ mmap->size + sizeof (mmap->size))) {
			if(mmap->type == 1)
				SxConSetTextColor(ecBrightGreen);
			else
				SxConSetTextColor(ecBrightMagenta);

			SxConPrint("   * base = 0x%x%x, length = 0x%x%x, type = 0x%x\n",
				mmap->base_addr_high, mmap->base_addr_low, mmap->length_high,
				mmap->length_low, mmap->type);
		}

		SxConSetTextColor(ecBrightWhite);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_DRIVEINFO)) {
		sx_multiboot_drive_info* di;

		SxConPrint(" * Installed Drives: %p, length %u Bytes\n", mbi->drives_addr, mbi->drives_length);

		for(di = (sx_multiboot_drive_info*)mbi->drives_addr
			; (sx_uint32)di < mbi->drives_addr + mbi->drives_length
			; di = (sx_multiboot_drive_info*)((sx_uint32)di + di->size)) {
			sx_uint16* ptr = &di->drive_ports;

			SxConPrint("    * num: 0x%x, mode: %s (%d), cyl.: %d, heads: %d, sect.: %d\n",
				di->drive_number, (di->drive_mode == 0 ? "CHS" : "LBA"), di->drive_mode,
				di->drive_cylinders, di->drive_heads, di->drive_sectors);
			
			while((sx_uint32)ptr < di->size) {
				if(*ptr == 0)
					break;

				SxConPrint("      * port: %x\n", *ptr++);
			}
		}
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_CONFIGTABLE)) {
		SxConPrint(" * Config Table: %x\n", mbi->config_table);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_LOADERNAME)) {
		SxConPrint(" * Bootloader: %s\n", mbi->loader_name);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_APMINFO)) {
		sx_multiboot_apminfo* ai = (sx_multiboot_apminfo*)mbi->apm_table;

		SxConPrint(" * APM Table: %p\n", ai);

		SxConPrint("    * Version: 0x%x, CSeg: 0x%x (%d Bytes), CSeg16: 0x%x (%d Bytes)\n",
			ai->apm_version, ai->apm_cseg, ai->apm_cseg_len, ai->apm_cseg16, ai->apm_cseg16_len);
		SxConPrint("      DSeg: 0x%x (%d Bytes), Entry Offset: 0x%x, Flags: 0x%x\n",
			ai->apm_dseg, ai->apm_dseg_len, ai->apm_offset, ai->apm_flags);
	}

	if(SX_BITTEST(mbi->flags, SX_BOOTFLAG_GRAPHICS)) {
		SxConPrint(" * VBE Information present.\n");

		SxConPrint("    * Control Info: 0x%x, Mode Info: 0x%x, Mode: 0x%x\n",
			mbi->vbe_control_info, mbi->vbe_mode_info, mbi->vbe_mode);
		SxConPrint("      Interface Seg.: 0x%x, Interface Off.: 0x%x, Interface Len.: %d Bytes\n",
			mbi->vbe_interface_seg, mbi->vbe_interface_off, mbi->vbe_interface_len);
	}

	SxCoreStatus("Booting ...");

	SxUnmaskAllPICInterupts();

	SxTimerInit(100);
	SxTimerTest();
}

void SxCoreStatus(const char* fmt, ...) {
	sx_eColors f, b;
	unsigned int height, x, y;
	sx_va_list va;
	sx_va_start(va, fmt);

	/* Get information about current state */
	SxConGetColor(&f, &b);
	SxConGetSize(SX_NULL, &height);
	SxConGetPos(&x, &y);

	/* Set Information for status line */
	SxConSetColor(ecBrightWhite, ecGreen);
	SxConSetPos(1, height);
	SxConClearLine();

	SxConPrintVa(fmt, va);

	SxConSetColor(f, b);
	SxConSetPos(x, y);

	sx_va_end(va);
}

void SxCorePanic(char const* fmt, ...) {
	sx_va_list va;
	sx_va_start(va, fmt);

	SxConSetTextColor(ecBrightRed);
	SxConPrintVa(fmt, va);

	sx_va_end(va);

	SxCoreHalt();
}

void SxCoreHalt() {
	SxCoreStatus("CPU Halted.");
	SX_CORE_IMMEDIATE_HALT;
}